@@ -28,13 +28,13 @@
 #include "kpassivepopup.h"
 #include "knotifyconfig.h"
 #include "knotification.h"
+#include "debug_p.h"
 
 #include <QBuffer>
 #include <QImage>
 #include <QLabel>
 #include <QTextDocument>
-#include <QApplication>
-#include <QDesktopWidget>
+#include <QGuiApplication>
 #include <QBoxLayout>
 #include <QLayout>
 #include <QDBusConnection>
@@ -43,15 +43,17 @@
 #include <QDBusError>
 #include <QDBusMessage>
 #include <QXmlStreamReader>
-#include <QDebug>
 #include <QMap>
 #include <QHash>
 #include <QXmlStreamEntityResolver>
 #include <QPointer>
 #include <QMutableListIterator>
+#include <QThread>
+#include <QScreen>
+#include <QFontMetrics>
+#include <QIcon>
 
 #include <kconfiggroup.h>
-#include <KIconThemes/KIconLoader>
 #include <KCodecs/KCharsets>
 
 static const char dbusServiceName[] = "org.freedesktop.Notifications";
@@ -66,12 +68,6 @@ public:
      * Fills the KPassivePopup with data
      */
     void fillPopup(KPassivePopup *popup, KNotification *notification, const KNotifyConfig &config);
-    /**
-     * Make sure a popup is completely supported by the notification backend.
-     * Changes the popup to be compatible if needed.
-     * @param notification the notification data to check
-     */
-    void ensurePopupCompatibility(KNotification *notification);
     /**
      * Removes HTML from a given string. Replaces line breaks with \n and
      * HTML entities by their 'normal forms'.
@@ -164,16 +160,14 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
     d->animationTimer = 0;
     d->dbusServiceExists = false;
     d->dbusServiceCapCacheDirty = true;
-
-    QRect screen = QApplication::desktop()->availableGeometry();
-    d->nextPosition = screen.top();
+    d->nextPosition = -1;
 
     // check if service already exists on plugin instantiation
     QDBusConnectionInterface *interface = QDBusConnection::sessionBus().interface();
     d->dbusServiceExists = interface && interface->isServiceRegistered(dbusServiceName);
 
     if (d->dbusServiceExists) {
-        onServiceOwnerChanged(dbusServiceName, QString(), "_"); //connect signals
+        onServiceOwnerChanged(dbusServiceName, QString(), QStringLiteral("_")); //connect signals
     }
 
     // to catch register/unregister events from service in runtime
@@ -190,10 +184,10 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
         startfdo = true;
 #else
         if (qgetenv("KDE_FULL_SESSION").isEmpty()) {
-            QDBusMessage message = QDBusMessage::createMethodCall("org.freedesktop.DBus",
-                                                                  "/org/freedesktop/DBus",
-                                                                  "org.freedesktop.DBus",
-                                                                  "ListActivatableNames");
+            QDBusMessage message = QDBusMessage::createMethodCall(QStringLiteral("org.freedesktop.DBus"),
+                                                                  QStringLiteral("/org/freedesktop/DBus"),
+                                                                  QStringLiteral("org.freedesktop.DBus"),
+                                                                  QStringLiteral("ListActivatableNames"));
 
             // FIXME - this should be async
             QDBusReply<QStringList> reply = QDBusConnection::sessionBus().call(message);
@@ -259,32 +253,46 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
 
     // if Growl can display our popups, use that instead
     if (NotifyByPopupGrowl::canPopup()) {
-        d->ensurePopupCompatibility(notification);
 
         QString appCaption, iconName;
         d->getAppCaptionAndIconName(notifyConfig, &appCaption, &iconName);
+        appCaption = d->stripHtml(appCaption);
 
         //did the user override the icon name?
         if (!notification->iconName().isEmpty()) {
             iconName = notification->iconName();
         }
 
-        KIconLoader iconLoader(iconName);
-        QPixmap appIcon = iconLoader.loadIcon(iconName, KIconLoader::Small);
+        // Our growl implementation does not support html stuff
+        // so strip it off right away
+        QString text = notification->text();
+        text = d->stripHtml(text);
 
-        NotifyByPopupGrowl::popup(&appIcon, timeout, appCaption, notification->text());
+        // The first arg is QPixmap*, however that pixmap is not used
+        // at all (it has Q_UNUSED) so just set it to 0
+        NotifyByPopupGrowl::popup(Q_NULLPTR, timeout, appCaption, text);
 
         // Finish immediately, because current NotifyByPopupGrowl can't callback
         finish(notification);
         return;
     }
 
+    // Check if this object lives in the GUI thread and return if it doesn't
+    // as Qt cannot create/handle widgets in non-GUI threads
+    if (QThread::currentThread() != qApp->thread()) {
+        qCWarning(LOG_KNOTIFICATIONS) << "KNotification did not detect any running org.freedesktop.Notifications server and fallback notifications cannot be used from non-GUI thread!";
+        return;
+    }
+
     // last fallback - display the popup using KPassivePopup
     KPassivePopup *pop = new KPassivePopup(notification->widget());
     d->passivePopups.insert(notification, pop);
     d->fillPopup(pop, notification, notifyConfig);
 
-    QRect screen = QApplication::desktop()->availableGeometry();
+    QRect screen = QGuiApplication::primaryScreen()->availableGeometry();
+    if (d->nextPosition == -1) {
+        d->nextPosition = screen.top();
+    }
     pop->setAutoDelete(true);
     connect(pop, SIGNAL(destroyed()), this, SLOT(onPassivePopupDestroyed()));
 
@@ -325,7 +333,7 @@ void NotifyByPopup::timerEvent(QTimerEvent *event)
     }
 
     bool cont = false;
-    QRect screen = QApplication::desktop()->availableGeometry();
+    QRect screen = QGuiApplication::primaryScreen()->availableGeometry();
     d->nextPosition = screen.top();
 
     Q_FOREACH (KPassivePopup *popup, d->passivePopups)
@@ -434,21 +442,21 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
         bool connected = QDBusConnection::sessionBus().connect(QString(), // from any service
                                                                dbusPath,
                                                                dbusInterfaceName,
-                                                               "ActionInvoked",
+                                                               QStringLiteral("ActionInvoked"),
                                                                this,
                                                                SLOT(onGalagoNotificationActionInvoked(uint,QString)));
         if (!connected) {
-            qWarning() << "warning: failed to connect to ActionInvoked dbus signal";
+            qCWarning(LOG_KNOTIFICATIONS) << "warning: failed to connect to ActionInvoked dbus signal";
         }
 
         connected = QDBusConnection::sessionBus().connect(QString(), // from any service
                                                           dbusPath,
                                                           dbusInterfaceName,
-                                                          "NotificationClosed",
+                                                          QStringLiteral("NotificationClosed"),
                                                           this,
                                                           SLOT(onGalagoNotificationClosed(uint,uint)));
         if (!connected) {
-            qWarning() << "warning: failed to connect to NotificationClosed dbus signal";
+            qCWarning(LOG_KNOTIFICATIONS) << "warning: failed to connect to NotificationClosed dbus signal";
         }
     }
 }
@@ -457,7 +465,6 @@ void NotifyByPopup::onGalagoNotificationActionInvoked(uint notificationId, const
 {
     auto iter = d->galagoNotifications.find(notificationId);
     if (iter == d->galagoNotifications.end()) {
-        qWarning() << "Failed to find KNotification id for dbus_id" << notificationId << "- action not triggered";
         return;
     }
 
@@ -474,10 +481,8 @@ void NotifyByPopup::onGalagoNotificationActionInvoked(uint notificationId, const
 
 void NotifyByPopup::onGalagoNotificationClosed(uint dbus_id, uint reason)
 {
-    Q_UNUSED(reason)
     auto iter = d->galagoNotifications.find(dbus_id);
     if (iter == d->galagoNotifications.end()) {
-        qWarning() << "Failed to find KNotification for dbus_id" << dbus_id;
         return;
     }
     KNotification *n = *iter;
@@ -489,8 +494,10 @@ void NotifyByPopup::onGalagoNotificationClosed(uint dbus_id, uint reason)
         // if the user closes the popup, it means he wants to get rid
         // of the notification completely, including playing sound etc
         // Therefore we close the KNotification completely after closing
-        // the popup
-        n->close();
+        // the popup, but only if the reason is 2, which means "user closed"
+        if (reason == 2) {
+            n->close();
+        }
     }
 }
 
@@ -500,7 +507,7 @@ void NotifyByPopup::onGalagoServerReply(QDBusPendingCallWatcher *watcher)
     watcher->deleteLater();
     KNotification *notification = watcher->property("notificationObject").value<KNotification*>();
     if (!notification) {
-        qWarning() << "Invalid notification object passed in DBus reply watcher; notification will probably break";
+        qCWarning(LOG_KNOTIFICATIONS) << "Invalid notification object passed in DBus reply watcher; notification will probably break";
         return;
     }
 
@@ -524,10 +531,10 @@ void NotifyByPopup::onGalagoServerCapabilitiesReceived(const QStringList &capabi
 
 void NotifyByPopupPrivate::getAppCaptionAndIconName(const KNotifyConfig &notifyConfig, QString *appCaption, QString *iconName)
 {
-    KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QString("Global"));
+    KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QStringLiteral("Global"));
     *appCaption = globalgroup.readEntry("Name", globalgroup.readEntry("Comment", notifyConfig.appname));
 
-    KConfigGroup eventGroup(&(*notifyConfig.eventsfile), QString("Event/%1").arg(notifyConfig.eventid));
+    KConfigGroup eventGroup(&(*notifyConfig.eventsfile), QStringLiteral("Event/%1").arg(notifyConfig.eventid));
     if (eventGroup.hasKey("IconName")) {
         *iconName = eventGroup.readEntry("IconName", notifyConfig.appname);
     } else {
@@ -547,10 +554,9 @@ void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notifi
     // capabilities immediately, so we don't need to wait for callback as in the case
     // of galago notifications
     queryPopupServerCapabilities();
-    ensurePopupCompatibility(notification);
 
-    KIconLoader iconLoader(iconName);
-    QPixmap appIcon = iconLoader.loadIcon(iconName, KIconLoader::Small);
+    int iconDimension = QFontMetrics(QFont()).height();
+    QPixmap appIcon = QIcon::fromTheme(iconName).pixmap(iconDimension, iconDimension);
 
     QWidget *vb = popup->standardView(notification->title().isEmpty() ? appCaption : notification->title(),
                                       notification->pixmap().isNull() ? notification->text() : QString(),
@@ -587,14 +593,14 @@ void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notifi
 
 
     if (!notification->actions().isEmpty()) {
-        QString linkCode = QString::fromLatin1("<p align=\"right\">");
+        QString linkCode = QStringLiteral("<p align=\"right\">");
         int i = 0;
         Q_FOREACH (const QString &it, notification->actions()) {
             i++;
-            linkCode += QString::fromLatin1("&nbsp;<a href=\"%1/%2\">%3</a>").arg(QString::number(notification->id()), QString::number(i), it.toHtmlEscaped());
+            linkCode += QStringLiteral("&nbsp;<a href=\"%1/%2\">%3</a>").arg(QString::number(notification->id()), QString::number(i), it.toHtmlEscaped());
         }
 
-        linkCode += QString::fromLatin1("</p>");
+        linkCode += QLatin1String("</p>");
         QLabel *link = new QLabel(linkCode , vb );
         link->setTextInteractionFlags(Qt::LinksAccessibleByMouse);
         link->setOpenExternalLinks(false);
@@ -618,7 +624,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         }
     }
 
-    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(dbusServiceName, dbusPath, dbusInterfaceName, "Notify");
+    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(dbusServiceName, dbusPath, dbusInterfaceName, QStringLiteral("Notify"));
 
     QList<QVariant> args;
 
@@ -631,25 +637,38 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         iconName = notification->iconName();
     }
 
-    // FIXME: rename this to something better reflecting what this is doing...maybe
-    ensurePopupCompatibility(notification);
-
     args.append(appCaption); // app_name
     args.append(updateId);  // notification to update
     args.append(iconName); // app_icon
-    args.append(notification->title().isEmpty() ? appCaption : notification->title()); // summary
-    args.append(notification->text()); // body
+
+    QString title = notification->title().isEmpty() ? appCaption : notification->title();
+    QString text = notification->text();
+
+    if (!popupServerCapabilities.contains(QStringLiteral("body-markup"))) {
+        if (title.startsWith(QLatin1String("<html>"))) {
+            title = stripHtml(title);
+        }
+        if (text.startsWith(QLatin1String("<html>"))) {
+            text = stripHtml(text);
+        }
+    }
+
+    args.append(title); // summary
+    args.append(text); // body
+
     // galago spec defines action list to be list like
     // (act_id1, action1, act_id2, action2, ...)
     //
     // assign id's to actions like it's done in fillPopup() method
     // (i.e. starting from 1)
     QStringList actionList;
-    int actId = 0;
-    Q_FOREACH (const QString &actionName, notification->actions()) {
-        actId++;
-        actionList.append(QString::number(actId));
-        actionList.append(actionName);
+    if (popupServerCapabilities.contains(QStringLiteral("actions"))) {
+        int actId = 0;
+        Q_FOREACH (const QString &actionName, notification->actions()) {
+            actId++;
+            actionList.append(QString::number(actId));
+            actionList.append(actionName);
+        }
     }
 
     args.append(actionList); // actions
@@ -659,7 +678,15 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
     // According to fdo spec, the app_name is supposed to be the applicaton's "pretty name"
     // but in some places it's handy to know the application name itself
     if (!notification->appName().isEmpty()) {
-        hintsMap["x-kde-appname"] = notification->appName();
+        hintsMap[QStringLiteral("x-kde-appname")] = notification->appName();
+    }
+
+    if (!notification->eventId().isEmpty()) {
+        hintsMap[QStringLiteral("x-kde-eventId")] = notification->eventId();
+    }
+
+    if (notification->flags() & KNotification::SkipGrouping) {
+        hintsMap["x-kde-skipGrouping"] = 1;
     }
 
     //FIXME - reenable/fix
@@ -670,7 +697,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         buffer.open(QIODevice::WriteOnly);
         notification->pixmap().save(&buffer, "PNG");
         buffer.close();
-        hintsMap["image_data"] = ImageConverter::variantForImage(QImage::fromData(pixmapData));
+        hintsMap[QStringLiteral("image_data")] = ImageConverter::variantForImage(QImage::fromData(pixmapData));
     }
 
     args.append(hintsMap); // hints
@@ -700,12 +727,12 @@ void NotifyByPopupPrivate::closeGalagoNotification(KNotification *notification)
     uint galagoId = galagoNotifications.key(notification, 0);
 
     if (galagoId == 0) {
-        qDebug() << "not found dbus id to close" << notification->id();
+        qCDebug(LOG_KNOTIFICATIONS) << "not found dbus id to close" << notification->id();
         return;
     }
 
     QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName, dbusPath,
-                                                    dbusInterfaceName, "CloseNotification");
+                                                    dbusInterfaceName, QStringLiteral("CloseNotification"));
     QList<QVariant> args;
     args.append(galagoId);
     m.setArguments(args);
@@ -714,7 +741,7 @@ void NotifyByPopupPrivate::closeGalagoNotification(KNotification *notification)
     bool queued = QDBusConnection::sessionBus().send(m);
 
     if (!queued) {
-        qWarning() << "Failed to queue dbus message for closing a notification";
+        qCWarning(LOG_KNOTIFICATIONS) << "Failed to queue dbus message for closing a notification";
     }
 
 }
@@ -726,8 +753,8 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
             popupServerCapabilities = NotifyByPopupGrowl::capabilities();
         } else {
             // Return capabilities of the KPassivePopup implementation
-            popupServerCapabilities = QStringList() << "actions" << "body" << "body-hyperlinks"
-                                                      << "body-markup" << "icon-static";
+            popupServerCapabilities = QStringList() << QStringLiteral("actions") << QStringLiteral("body") << QStringLiteral("body-hyperlinks")
+                                                      << QStringLiteral("body-markup") << QStringLiteral("icon-static");
         }
     }
 
@@ -735,7 +762,7 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
         QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName,
                                                         dbusPath,
                                                         dbusInterfaceName,
-                                                        "GetCapabilities");
+                                                        QStringLiteral("GetCapabilities"));
 
         QDBusConnection::sessionBus().callWithCallback(m,
                                                        q,
@@ -745,22 +772,6 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
     }
 }
 
-void NotifyByPopupPrivate::ensurePopupCompatibility(KNotification *notification)
-{
-    if (!popupServerCapabilities.contains("actions")) {
-        notification->setActions(QStringList());
-    }
-
-    if (!popupServerCapabilities.contains("body-markup")) {
-        if (notification->title().startsWith("<html>")) {
-            notification->setTitle(stripHtml(notification->title()));
-        }
-        if (notification->text().startsWith("<html>")) {
-            notification->setText(stripHtml(notification->text()));
-        }
-    }
-}
-
 QString NotifyByPopupPrivate::stripHtml(const QString &text)
 {
     QXmlStreamReader r("<elem>" + text + "</elem>");
@@ -778,7 +789,7 @@ QString NotifyByPopupPrivate::stripHtml(const QString &text)
 
     if (r.hasError()) {
         // XML error in the given text, just return the original string
-        qWarning() << "Notification to send to backend which does "
+        qCWarning(LOG_KNOTIFICATIONS) << "Notification to send to backend which does "
                          "not support HTML, contains invalid XML:"
                       << r.errorString() << "line" << r.lineNumber()
                       << "col" << r.columnNumber();
@@ -799,7 +810,7 @@ QString NotifyByPopupPrivate::HtmlEntityResolver::resolveUndeclaredEntity(const
     QChar ent = KCharsets::fromEntity('&' + name);
 
     if (ent.isNull()) {
-        qWarning() << "Notification to send to backend which does "
+        qCWarning(LOG_KNOTIFICATIONS) << "Notification to send to backend which does "
                       "not support HTML, contains invalid entity: "
                    << name;
         ent = ' ';
